/* file:  es10_circolareA3_vet3cond.c 
   Routine che fornisce tre passaggi di testimone tra 3 processi
   il primo fa partire il secondo, il secondo fa partire il terzo
   e il terzo fa partire nuovamente il primo e cosi via
   Implementata con una condition variable per ogni pthread.
*/ 

#ifndef _THREAD_SAFE
	#define _THREAD_SAFE
#endif
#ifndef _POSIX_C_SOURCE
	#define _POSIX_C_SOURCE 200112L
#else
#if _POSIX_C_SOURCE < 200112L
	#undef  _POSIX_C_SOURCE
	#define _POSIX_C_SOURCE 200112L
#endif
#endif


#include <unistd.h> 
#include <stdlib.h> 
#include <stdio.h> 
#include <stdint.h>
#include <inttypes.h>
#include <pthread.h> 

#include "printerror.h"
#include "DBGpthread.h"

#define NUMTHREADS 3

/* dati da proteggere */
uint64_t valGlobale=0;

/* variabili per la sincronizzazione */
pthread_mutex_t  mutex; 
pthread_cond_t  Vcond[NUMTHREADS]; 
int FaiWait[NUMTHREADS];

void *ThreadFunc (void *arg) 
{ 
	intptr_t indice; char Label[100];

	indice=(intptr_t)arg;
	sprintf(Label, "T%lu"PRIiPTR , (unsigned long int) indice );
	while(1) {
		/* aspettiamo che il precedente ci risvegli */
		DBGpthread_mutex_lock(&mutex,Label); 

		/* se il precedente ha gia' fatto la signal allora
		   il nostro deve proseguire senza fare la wait
		   */
		if( FaiWait[indice] ) {
			DBGpthread_cond_wait(&Vcond[indice],&mutex,Label);
		}
		/* al prossimo giro devo fare la wait, 
		   a meno che il mio precedente non faccia la signal
		   prima che io tenti di fare la wait
		*/
		FaiWait[indice]=1;

		/* SEZIONE CRITICA: legge cio' che e' stato messo da Primo e lo cambia */
		valGlobale=valGlobale+15;

		printf("%" PRIiPTR " scrive %lu \n", indice, valGlobale ); 
		fflush(stdout);

		/* risveglio il successivo */
		DBGpthread_cond_signal(&Vcond[ (indice+1)%NUMTHREADS ], Label); 
		/* Nel caso che Primo non abbia ancora fatto la wait
		   allora Terzo dice a Primo che non deve aspettare
		   perche' Terzo ha gia' fatto la signal
		   e cosi' per gli altri
		*/
		FaiWait[(indice+1)%NUMTHREADS]=0;

		/* rilascio mutua esclusione, cosi' il successivo parte */
		DBGpthread_mutex_unlock(&mutex,Label); 
	}
	printf("%" PRIiPTR " termina\n", indice); 
	fflush(stdout);
	pthread_exit(NULL); 
}

 
int main () 
{ 
	pthread_t    th[NUMTHREADS]; 
	int  rc; intptr_t i;
	void *ptr; 

	rc = pthread_mutex_init(&mutex, NULL);
	if( rc ) PrintERROR_andExit(rc,"pthread_mutex_init failed");

	for(i=0;i<NUMTHREADS;i++) {
		rc = pthread_cond_init(&(Vcond[i]), NULL);
		if( rc ) PrintERROR_andExit(rc,"pthread_cond_init failed");
	}
	FaiWait[0]=0;
	FaiWait[1]=1;
	FaiWait[2]=1;
	for(i=0;i<NUMTHREADS;i++) {
		rc = pthread_create(&(th[i]), NULL, ThreadFunc, (void*)i ); 
		if (rc) PrintERROR_andExit(rc,"pthread_create failed");
	}

	for(i=0;i<NUMTHREADS;i++) {
		rc = pthread_join(th[i], &ptr ); 
		if (rc) PrintERROR_andExit(rc,"pthread_join failed");
	}

	rc = pthread_mutex_destroy(&mutex); 
	if(rc) PrintERROR_andExit(rc,"pthread_mutex_destroy failed");
	for(i=0;i<NUMTHREADS;i++) {
		rc = pthread_cond_destroy(&(Vcond[i])); 
		if(rc) PrintERROR_andExit(rc,"pthread_cond_destroy failed");
	}

	return(0); 
} 
  
  
